__version__ = '0.2.1'

import os
import copy
import yaml
import numpy as np
import scipy.linalg as la
from huckelpy import tools, atomic_orbital
from huckelpy.overlap_integral import overlap_integral


class ExtendedHuckel:

    def __init__(self, coordinates, symbols, charge=0, cartesian_orbitals=False):

        self.coordinates = coordinates
        self.symbols = symbols
        self.charge = charge
        self.pure_orbitals = not cartesian_orbitals
        self.n_electrons = 0
        with open(os.path.dirname(os.path.abspath(__file__)) + '/basis_set.yaml') as file:
            self.basis = yaml.load(file, Loader=yaml.FullLoader)
        self.overlap = np.zeros((self.matrices_dimensions(), self.matrices_dimensions()))
        self.hamiltonian = None
        self.eigenvalues = None
        self.eigenvectors = None
        self.total_energy = None
        self.coefficients = []
        self.molecular_basis = None
        self.orbitals_list = []
        self.orbital_vector = []
        self.orbitals_atom = []
        self.atomic_numbers = []

    @property
    def multiplicity(self):
        return self.get_multiplicity()

    @property
    def number_of_electrons(self):
        return self.get_number_of_electrons()

    def set_atom_energy(self, symbol, sp_energy, d_energy=None, f_energy=None):

        self.basis[symbol][0]['VSIP'] = sp_energy
        if d_energy is not None:
            self.basis[symbol][1]['VSIP'] = d_energy
        if f_energy is not None:
            self.basis[symbol][2]['VSIP'] = f_energy
        self.overlap = None
        self.hamiltonian = None
        self.eigenvalues = None
        self.eigenvectors = None

    def get_coordinates(self):
        return self.coordinates

    def get_symbols(self):
        return self.symbols

    def get_basis(self):
        return self.basis

    def get_overlap_matrix(self):
        if not np.any(self.overlap):
            self.overlap_matrix()
        return self.overlap

    def get_hamiltonian_matrix(self):
        if self.hamiltonian is None:
            self.hamiltonian_matrix()
        return self.hamiltonian

    def get_mo_coefficients(self):
        if not self.coefficients:
            self.norm_coefficients()
        return self.coefficients

    def norm_coefficients(self):
        for mo_coefficients in self.get_eigenvectors():
            norm = 0
            for coef in mo_coefficients:
                norm += coef**2
            norm = 1 / np.sqrt(norm)
            self.coefficients.append(norm * mo_coefficients)

    def get_mo_energies(self):
        if self.eigenvalues is None:
            self.calculate_energy()
        return self.eigenvalues

    def get_eigenvectors(self):
        if self.eigenvectors is None:
            self.calculate_energy()
        return self.eigenvectors

    def get_total_energy(self):
        if self.total_energy is None:
            self.total_energy = [ei * 2 for ei in self.get_mo_energies()[:int((self.get_number_of_electrons()) / 2)]]
            if self.get_number_of_electrons() % 2 == 1:
                self.total_energy += self.get_mo_energies()[int((self.get_number_of_electrons()) / 2)]
            self.total_energy = sum(self.total_energy)
        return self.total_energy

    def matrices_dimensions(self):
        matrices_dimensions = 0
        for symbol in self.symbols:
            atomic_orbitals = 0
            for shell in self.basis[symbol]:
                if shell['shell_type'] == 'sp':
                    atomic_orbitals += 4
                elif shell['shell_type'] == 'd':
                    if self.pure_orbitals:
                        atomic_orbitals += 5
                    else:
                        atomic_orbitals += 6
                elif shell['shell_type'] == 'f':
                    if self.pure_orbitals:
                        atomic_orbitals += 7
                    else:
                        atomic_orbitals += 10
                else:
                    atomic_orbitals += 1
            matrices_dimensions += atomic_orbitals
        return matrices_dimensions

    def calculate_energy(self):
        self.eigenvalues, self.eigenvectors = la.eigh(self.get_hamiltonian_matrix(), b=self.get_overlap_matrix())
        idx = np.argsort(self.eigenvalues)
        self.eigenvalues = self.eigenvalues[idx]
        self.eigenvectors = self.eigenvectors[:, idx].T

    def orbitalxorbital(self, orbital_obj1, orbital_obj2):

        g0 = 0.
        for ido1, orbital1 in enumerate(orbital_obj1.ao_linear_combination[0]):
            for ido2, orbital2 in enumerate(orbital_obj2.ao_linear_combination[0]):
                g1g2 = 0
                for g1_parameters in range(3):
                    for g2_parameters in range(3):
                        g1g2 += orbital_obj1.ao_coefficients[g1_parameters]*orbital_obj2.ao_coefficients[g2_parameters] * \
                                orbital_obj1.gaussian_norm[g1_parameters]*orbital_obj2.gaussian_norm[g2_parameters] * \
                                orbital_obj1.real_spherical_harmonic() * orbital_obj2.real_spherical_harmonic() * \
                                overlap_integral(orbital1, orbital2, orbital_obj1.position, orbital_obj2.position,
                                                 orbital_obj1.ao_exponents[g1_parameters],
                                                 orbital_obj2.ao_exponents[g2_parameters])
                g0 += orbital_obj1.ao_linear_combination[1][ido1]*orbital_obj2.ao_linear_combination[1][ido2]*g1g2

        return g0

    def return_shell_position(self, shell_list, type):
        for idsh, shell_type in enumerate(shell_list):
            if type in shell_type['shell_type']:
                return idsh

    def overlap_matrix(self):

        for ido1, orbital_obj1 in enumerate(self.get_ao_list()):
            for ido2, orbital_obj2 in enumerate(self.get_ao_list()[ido1:]):
                self.overlap[ido1][ido1+ido2] = self.orbitalxorbital(orbital_obj1, orbital_obj2)

        self.overlap = self.overlap + self.overlap.T - np.diag(np.diag(self.overlap))

        norm_vector = [1 / np.sqrt(x) for x in np.diag(self.overlap)]
        for i in range(len(self.overlap)):
            for j in range(len(self.overlap[i])):
                self.overlap[i][j] *= norm_vector[i] * norm_vector[j]

    def hamiltonian_matrix(self):

        self.hamiltonian = np.zeros((self.matrices_dimensions(), self.matrices_dimensions()))
        ev_to_ha = 3.67493095E-2
        k = 0.75
        if not np.any(self.overlap):
            self.get_overlap_matrix()
        for ido1, orbital_obj1 in enumerate(self.get_ao_list()):
            for ido2, orbital_obj2 in enumerate(self.get_ao_list()):
                if ido1 == ido2:
                    h = orbital_obj1.energy
                    self.hamiltonian[ido1][ido2] = h * ev_to_ha
                else:
                    h1 = orbital_obj1.energy
                    h2 = orbital_obj2.energy
                    a = (h1 - h2) / (h1 + h2)
                    eh_constant = 1 + k + a ** 2 - k * a ** 4
                    self.hamiltonian[ido1][ido2] = eh_constant * (h1 + h2) * (1 / 2) * ev_to_ha * self.overlap[ido1][ido2]

    def get_ao_list(self):
        cartesian_basis = {'s': ['s'],
                           'sp': ['s', 'px', 'py', 'pz'],
                           'd': ['dxx', 'dyy', 'dzz', 'dxy', 'dxz', 'dyz'],
                           'f': ['fxxx', 'fyyy', 'fzzz', 'fxyy', 'fxxy', 'fxxz', 'fxzz', 'fyzz', 'fyyz', 'fxyz']}
        pure_basis = {'s': ['s'],
                      'sp': ['s', 'px', 'py', 'pz'],
                      'd': ['dz2', 'dxz', 'dyz', 'dx2-y2', 'dxy']}

        if self.pure_orbitals:
            basis = pure_basis
        else:
            basis = cartesian_basis

        if not self.orbitals_list:
            for ids, symbol in enumerate(self.symbols):
                for shell in self.basis[symbol]:
                    for orbital_type in basis[shell['shell_type']]:
                        if 'p' in orbital_type:
                            eh_energy = shell['VSIP'][1]
                        else:
                            eh_energy = shell['VSIP'][0]
                        self.orbitals_list.append([atomic_orbital.AtomicOrbital(symbol, shell, self.coordinates[ids],
                                                                                orbital_type, eh_energy)])
            self.orbitals_list = [y for x in self.orbitals_list for y in x]
        return self.orbitals_list

    def get_number_of_electrons(self):
        if self.n_electrons == 0:
            for symbol in self.symbols:
                symbol = symbol.split('_')[0]
                self.n_electrons += tools.element_valence_electron(symbol)
            self.n_electrons -= self.charge
        return self.n_electrons

    def get_multiplicity(self):
        if self.get_number_of_electrons() % 2 == 0:
            self._multiplicity = 1
        else:
            self._multiplicity = 2
        return self._multiplicity

    def get_atomic_numbers(self):
        if not self.atomic_numbers:
            for symbol in self.symbols:
                symbol = symbol.split('_')[0]
                self.atomic_numbers.append(tools.element_to_atomic_number(symbol))
        return self.atomic_numbers

    def get_charge(self):
        return self.charge

    def get_molecular_basis(self):
        if self.molecular_basis is None:
            self.molecular_basis = {'name': 'STO-3G',
                                    'primitive_type': 'gaussian',
                                    'atoms': []}
            for symbol in self.symbols:
                pure_basis = copy.deepcopy(self.basis[symbol])
                if self.pure_orbitals and len(self.basis[symbol]) > 1:
                    pure_basis[1]['shell_type'] = 'd_'
                self.molecular_basis['atoms'].append({'shells': pure_basis})
        return self.molecular_basis



import numpy as np

eV_to_Ha = 3.67493095E-2


def build_output(parsed_data, filename):

    output = open(filename+'.out', 'w')

    structure = np.array(parsed_data.get_coordinates())  # ['structure']
    symbols = parsed_data.get_symbols()
    alpha_mo_coeff = parsed_data.get_eigenvectors()  # ['coefficients']['alpha']
    # alpha_mo_coeff = parsed_data.get_mo_coefficients()
    alpha_mo_energies = np.real(parsed_data.get_mo_energies())  # ['mo_energies']['alpha']
    ao_list = parsed_data.get_ao_list()
    overlap = parsed_data.get_overlap_matrix()
    hamiltonian = parsed_data.get_hamiltonian_matrix()

    output.write('Extended Huckel calculation\n')
    output.write('Current cartesian coordinates:\n')
    for ids, symbol in enumerate(symbols):
        output.write('{:4} {:18.8f} {:18.8f} {:18.8f}'.format(symbol,
                                                              structure[ids][0],
                                                              structure[ids][1],
                                                              structure[ids][2]))
        output.write('\n')

    output.write('Molecular Orbital Coefficients: \n')
    for i in range(int(len(alpha_mo_energies)/5) + (len(alpha_mo_energies)%5 > 0)):

        n_orbitals = list(range(5*i, 5*i + 5))
        if i == int(len(alpha_mo_energies)/5):
            if len(alpha_mo_energies)%5 > 0:
                n_orbitals = n_orbitals[:len(alpha_mo_energies[n_orbitals[0]:])]

        width = 20
        for orbital in n_orbitals:
            output.write('{:{width}}'.format(orbital + 1, width=width))
            width = 18
        output.write('\n')
        width = 14
        output.write('Eigenvalues')
        for orbital in n_orbitals:
            output.write('{:{width}.8f}'.format(alpha_mo_energies[orbital]/eV_to_Ha, width=width))
            width = 18
        output.write('\n')
        output.write('\n')

        for j in range(len(alpha_mo_coeff[i])):
            width = 15
            output.write('{:4} {:5}'.format(j+1, ao_list[j].row+ao_list[j].type))
            if ao_list[j].type == 'dx2-y2':
                width = 13

            for orbital in n_orbitals:
                output.write('{:{width}.8f}'.format(
                    alpha_mo_coeff[orbital][j], width=width))
                width = 18
            output.write('\n')

    output.write('\n')
    output.write('Total Energy = {:18.8f}\n'.format(parsed_data.get_total_energy()/eV_to_Ha))
    output.write('\n')

    # La llargada dels mo no es la mateixa que el solapament i l'hamiltonià
    output.write('Overlap matrix: \n')
    for i in range(int(len(alpha_mo_energies)/5) + (len(alpha_mo_energies)%5 > 0)):

        n_orbitals = list(range(5*i, 5*i + 5))
        if i == int(len(alpha_mo_energies)/5):
            if len(alpha_mo_energies)%5 > 0:
                n_orbitals = n_orbitals[:len(alpha_mo_energies[n_orbitals[0]:])]

        width = 20
        for orbital in n_orbitals:
            output.write('{:{width}}'.format(orbital + 1, width=width))
            width = 18
        output.write('\n')
        output.write('\n')

        for j in range(len(alpha_mo_coeff[i])):
            output.write('{:3}'.format(j+1))
            width = 21
            for orbital in n_orbitals:
                output.write('{:{width}.8f}'.format(
                    overlap[orbital][j], width=width))
                width = 18
            output.write('\n')

    output.write('\n')

    output.write('Hamiltonian matrix: \n')
    for i in range(int(len(alpha_mo_energies)/5) + (len(alpha_mo_energies)%5 > 0)):

        n_orbitals = list(range(5*i, 5*i + 5))
        if i == int(len(alpha_mo_energies)/5):
            if len(alpha_mo_energies)%5 > 0:
                n_orbitals = n_orbitals[:len(alpha_mo_energies[n_orbitals[0]:])]

        width = 20
        for orbital in n_orbitals:
            output.write('{:{width}}'.format(orbital + 1, width=width))
            width = 18
        output.write('\n')
        output.write('\n')

        for j in range(len(alpha_mo_coeff[i])):
            output.write('{:3}'.format(j+1))
            width = 21
            for orbital in n_orbitals:
                output.write('{:{width}.8f}'.format(
                    hamiltonian[orbital][j], width=width))
                width = 18
            output.write('\n')

    output.write('\n')


def get_array_txt(label, type, array, row_size=5):

    formats = {'R': '15.8e',
               'I': '11'}

    n_elements = len(array)
    rows = int(np.ceil(n_elements/row_size))

    txt_fchk = '{:40}   {}   N=       {:5}\n'.format(label, type, n_elements)

    for i in range(rows):
        if (i+1)*row_size > n_elements:
            txt_fchk += (' {:{fmt}}'* (n_elements - i*row_size) + '\n').format(*array[i * row_size:n_elements],
                                                                               fmt=formats[type])
        else:
            txt_fchk += (' {:{fmt}}'* row_size  + '\n').format(*array[i * row_size: (i+1)*row_size],
                                                               fmt=formats[type])

    return txt_fchk


def build_fchk(parsed_data):

    structure = np.array(parsed_data.get_coordinates())#['structure']
    basis = parsed_data.get_molecular_basis()#['basis']
    alpha_mo_coeff = parsed_data.get_eigenvectors()#['coefficients']['alpha']
    alpha_mo_energies = parsed_data.get_mo_energies()#['mo_energies']['alpha']

    #overlap = parsed_data['overlap']
    #coor_shell = parsed_data['coor_shell']
    #core_hamiltonian = parsed_data['core_hamiltonian']
    #scf_density = parsed_data['scf_density']

    number_of_basis_functions = len(alpha_mo_coeff)
    number_of_electrons = parsed_data.get_number_of_electrons()
    # number_of_electrons = np.sum(structure.get_atomic_numbers()) - structure.charge
    alpha_electrons = number_of_electrons//2
    beta_electrons = number_of_electrons//2
    if parsed_data.get_multiplicity() > 1:
        alpha_electrons += 1

    #print(alpha_electrons)
    #print(number_of_electrons)

    alpha_mo_coeff = np.array(alpha_mo_coeff).flatten().tolist()

    # if 'beta' in parsed_data.get_mo_coefficients(): #['coefficients']:
    #     beta_mo_coeff = parsed_data['coefficients']['beta']
    #     beta_mo_coeff = np.array(beta_mo_coeff).flatten().tolist()
    #
    #     beta_mo_energies = parsed_data['alpha_mo_energies']['beta']

    shell_type_list = {'s':  {'type':  0, 'angular_momentum': 0},
                       'p':  {'type':  1, 'angular_momentum': 1},
                       'd':  {'type':  2, 'angular_momentum': 2},
                       'f':  {'type':  3, 'angular_momentum': 3},
                       'sp': {'type': -1, 'angular_momentum': 1},  # hybrid
                       'd_': {'type': -2, 'angular_momentum': 2},  # pure
                       'f_': {'type': -3, 'angular_momentum': 3}}  # pure

    shell_type = []
    p_exponents = []
    c_coefficients = []
    p_c_coefficients = []
    n_primitives = []
    atom_map = []

    largest_degree_of_contraction = 0
    highest_angular_momentum = 0
    number_of_contracted_shells = 0

    for i, atoms in enumerate(basis['atoms']):
        for shell in atoms['shells']:
            number_of_contracted_shells += 1
            st = shell['shell_type']
            shell_type.append(shell_type_list[st]['type'])
            n_primitives.append(len(shell['p_exponents']))
            atom_map.append(i+1)
            if highest_angular_momentum < shell_type_list[st]['angular_momentum']:
                highest_angular_momentum = shell_type_list[st]['angular_momentum']

            if len(shell['con_coefficients']) > largest_degree_of_contraction:
                    largest_degree_of_contraction = len(shell['con_coefficients'])

            for p in shell['p_exponents']:
                p_exponents.append(p)
            for c in shell['con_coefficients']:
                c_coefficients.append(c)
            for pc in shell['p_con_coefficients']:
                p_c_coefficients.append(pc)

    angstrom_to_bohr = 1/0.529177249
    coordinates_list = angstrom_to_bohr*structure.flatten()

    txt_fchk = '{}\n'.format('Extended Huckel calculation over filename')
    txt_fchk += 'SP        RHuckel                                                     {}\n'.format('STO-3G')
    txt_fchk += 'Number of atoms                            I               {}\n'.format(len(structure))
    txt_fchk += 'Charge                                     I               {}\n'.format(parsed_data.get_charge())
    txt_fchk += 'Multiplicity                               I               {}\n'.format(parsed_data.get_multiplicity())
    txt_fchk += 'Number of electrons                        I               {}\n'.format(number_of_electrons)
    txt_fchk += 'Number of alpha electrons                  I               {}\n'.format(alpha_electrons)
    txt_fchk += 'Number of beta electrons                   I               {}\n'.format(beta_electrons)
    txt_fchk += 'Number of basis functions                  I               {}\n'.format(number_of_basis_functions)

    txt_fchk += get_array_txt('Atomic numbers', 'I', parsed_data.get_atomic_numbers(), row_size=6)
    txt_fchk += get_array_txt('Nuclear charges', 'R', parsed_data.get_atomic_numbers())
    txt_fchk += get_array_txt('Current cartesian coordinates', 'R', coordinates_list)

    txt_fchk += 'Number of contracted shells                I               {}\n'.format(number_of_contracted_shells)
    txt_fchk += 'Number of primitive shells                 I               {}\n'.format(np.sum(n_primitives))
    txt_fchk += 'Highest angular momentum                   I               {}\n'.format(highest_angular_momentum)
    txt_fchk += 'Largest degree of contraction              I               {}\n'.format(largest_degree_of_contraction)

    txt_fchk += get_array_txt('Shell types', 'I', shell_type, row_size=6)
    txt_fchk += get_array_txt('Number of primitives per shell', 'I', n_primitives, row_size=6)
    txt_fchk += get_array_txt('Shell to atom map', 'I', atom_map, row_size=6)
    txt_fchk += get_array_txt('Primitive exponents', 'R', p_exponents)
    txt_fchk += get_array_txt('Contraction coefficients', 'R', c_coefficients)
    txt_fchk += get_array_txt('P(S=P) Contraction coefficients', 'R', p_c_coefficients)
    # txt_fchk += get_array_txt('Coordinates of each shell', 'R', coor_shell) #
    # txt_fchk += get_array_txt('Overlap Matrix', 'R', overlap)
    #txt_fchk += get_array_txt('Core Hamiltonian Matrix', 'R', core_hamiltonian)
    txt_fchk += 'Total Energy                               R             {}\n'.format(parsed_data.get_total_energy())
    txt_fchk += get_array_txt('Alpha Orbital Energies', 'R', alpha_mo_energies)
    # txt_fchk += get_array_txt('Beta Orbital Energies', 'R', beta_mo_energies)
    # txt_fchk += get_array_txt('Total SCF Density', 'R', scf_density)
    txt_fchk += get_array_txt('Alpha MO coefficients', 'R', alpha_mo_coeff)
    # txt_fchk += get_array_txt('Beta MO coefficients', 'R', beta_mo_coeff)

    return txt_fchk


from numpy.distutils.core import setup


def get_version_number():
    for l in open('huckelpy/__init__.py', 'r').readlines():
        if not(l.find('__version__')):
            exec(l, globals())
            return __version__


setup(name='huckelpy',
      version=get_version_number(),
      description='Package to calculate the molecular orbitals with the extended huckel method',
      author='Efrem Bernuz',
      author_email='komuisan@gmail.com',
      packages=['huckelpy'],
      package_data={'': ['basis_set.yaml']},
      include_package_data=True,
      install_requires=['numpy', 'PyYAML', 'scipy'])